# -*- coding: utf-8 -*-

import json
import redis

from .base import BaseStatusWrapper

_HOST = 'localhost'
_PORT = '6379'
_DB_ID = 0

_STATUS_KEY = "okq_status:{name}"


class StatusWrapper(BaseStatusWrapper):
    """A wrapper for redis connection and custom methods"""

    def create_cnx(self, db_settings):
        return redis.StrictRedis(host=db_settings.get('HOST', _HOST),
                                 port=db_settings.get('PORT', _PORT),
                                 db=db_settings.get('DB_ID', _DB_ID))

    def db_init(self):
        pass

    def set_status(self, name, data):
        """Save the status in a redis list.

        Data is serialized in json
        """
        self._cnx.set(_STATUS_KEY.format(name=name), json.dumps(data))

    def get_status(self, name):
        """return the redis key with the given name and arguments.

        :param name: the name as found in `settings.SOURCES`
        :type name: str
        """
        return json.loads(self._cnx.get(_STATUS_KEY.format(name=name)))



# -*- coding: utf-8 -*-

from datetime import timedelta

from .source import get_sources


def get_schedule(seconds):
    """Get celery schedule.
    """
    dyn_schedule = dict()
    for source in get_sources():
        dyn_schedule[source['name']] = {
            'task': 'openkongqi.tasks.fetch',
            'schedule': timedelta(seconds=seconds),
            'args': (source['name'], )
        }
    return dyn_schedule



# -*- coding: utf-8 -*-

from importlib import import_module

from .conf import settings
from .exceptions import SourceError

from celery import Celery
from celery.utils.log import get_task_logger


app = Celery('okq')
app.config_from_object('celeryconfig')
logger = get_task_logger(__name__)


@app.task
def fetch(name):
    if name not in settings['SOURCES']:
        raise SourceError("Unknown source ({})".format(name))
    modname = settings['SOURCES'][name]['modname']
    mod = import_module(modname)
    src = mod.Source(name)
    src.scrape()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, print_function, unicode_literals
from copy import copy
import pytz

from sqlalchemy import create_engine
from sqlalchemy import Column, String, DateTime, Float
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base

from .base import BaseRecordsWrapper
from ..utils import get_uuid

Base = declarative_base()


class SQLAlchemyRecordsWrapper(BaseRecordsWrapper):

    def __init__(self, settings, cache, *args, **kwargs):
        self._engine = create_engine(self.create_dsn(settings))
        super(SQLAlchemyRecordsWrapper, self).__init__(
            settings, cache, *args, **kwargs
        )

    def create_dsn(self, settings):
        """Create a data source name (DNS) given a settings dict.

        .. warning:: This method has to be overwritten

        :returns: sqlalchemy.engine.url.URL instance
        """
        raise NotImplementedError

    def create_cnx(self, settings):
        db_session = scoped_session(sessionmaker(autocommit=False,
                                                 autoflush=False,
                                                 bind=self.get_engine()))
        return db_session

    def db_init(self):
        Base.metadata.create_all(bind=self.get_engine())

    def get_engine(self):
        """Return engine url / data source name (DSN) of database."""
        return self._engine

    def _get_rec_uuid(self, uuid, context=None):
        """Return contextualized station uuid.

        If there exists context with module name, the module name will simply
        be appended in the front for inserting into the database.
        """
        if context is not None:
            moduuid = context.get('moduuid')
            if moduuid:
                return get_uuid(moduuid, uuid)
        return uuid

    def is_duplicate(self, record):
        dup_count = self._cnx.query(Record).filter_by(
            ts=record.ts, uuid=record.uuid, key=record.key
        ).count()
        return dup_count != 0

    def write_records(self, records, ignore_check_latest=False, context=None):
        for uuid, records in records.items():
            rec_uuid = self._get_rec_uuid(uuid, context=context)
            rows = []
            latest = self.get_latest(uuid, context=context)
            # this is a very naive checking of which records to consider
            # when inserting the databse because it simply limits records
            # for those that are later than the latest timestamp
            # however, if we somehow want to manually insert into database
            # records that are before timestamp on `get_latest`
            # none of the records will be inserted
            if not ignore_check_latest and latest is not None:
                records = filter(lambda r: r['ts'] > latest['ts'], records)
            last_record = copy(latest)
            for record in records:
                ts = record['ts'].astimezone(pytz.utc)
                # update last record as we go
                if last_record is None:
                    last_record = record
                else:
                    if ts > last_record['ts']:
                        last_record = record
                # insert the records into SQL database
                for fieldname, value in record['fields'].items():
                    row = Record(ts=ts,
                                 uuid=rec_uuid,
                                 key=fieldname,
                                 value=value)
                    if not self.is_duplicate(row):
                        rows.append(row)
            self._cnx.add_all(rows)
            self._cnx.commit()
            # set latest cache value
            if last_record != latest:
                self.set_latest(uuid, last_record, context=context)

    def get_records(self, uuid, start, end, fields=None, context=None):
        # sanitize datetime input
        start_dt = to_utc(start)
        end_dt = to_utc(end)

        rec_uuid = self._get_rec_uuid(uuid, context=context)

        query = self._cnx.query(Record) \
            .filter(Record.uuid == rec_uuid) \
            .filter(Record.ts >= start_dt) \
            .filter(Record.ts <= end_dt)

        # filter which fields
        if fields is not None:
            query = query.filter(Record.key.in_(fields))

        distinct_datetimes = [
            q.ts for q in
            query.group_by('ts')
        ]

        # group by datetime
        for ts in distinct_datetimes:
            yield {
                'ts': ts.replace(tzinfo=pytz.utc),
                'fields': {
                    rec.key: rec.value
                    for rec
                    in query.filter(Record.ts == ts).all()
                },
            }


def to_utc(dt):
    """Make input time parameters UTC or force it."""
    if dt.tzinfo is None:
        return dt.replace(tzinfo=pytz.utc)
    else:
        return dt.astimezone(pytz.utc)


class Record(Base):
    __tablename__ = 'records'

    ts = Column(DateTime, primary_key=True)
    uuid = Column(String(250), nullable=False, primary_key=True)
    key = Column(String(250), nullable=False, primary_key=True)
    value = Column(Float(), nullable=True)

    def __repr__(self):
        return (
            "<Record(ts='{ts}', uuid='{uuid}', key='{key}', value='{value}')>"
            .format(ts=self.ts, uuid=self.uuid, key=self.key, value=self.value)
        )



# -*- coding: utf-8 -*-
import argparse
import distutils.spawn
import os
import sys

from openkongqi.exceptions import OpenKongqiError

from celery.bin.celery import main as clry_main


def okq_init():
    parser = argparse.ArgumentParser(
        description="initialize database for records"
    )
    parser.add_argument('--okqconf', dest='conffile', action='store',
                        type=str, help='path to a configuration file')
    args = parser.parse_args()

    if args.conffile is not None:
        os.environ.setdefault('OKQ_CONFMODULE', args.conffile)

    # run magic configuration after environment variable is set
    import openkongqi.conf
    openkongqi.conf.recsdb.db_init()


def okq_server():
    parser = argparse.ArgumentParser(
        description='outdoor air quality data',
        epilog='Any extra argument is directly passed to celery')
    parser.add_argument('--okqconf', dest='conffile', action='store',
                        type=str, help='path to a configuration file')
    args, clry_args = parser.parse_known_args()

    if args.conffile is not None:
        os.environ.setdefault('OKQ_CONFMODULE', args.conffile)

    # reproduce sys.argv to call celery
    clry_args.insert(0, distutils.spawn.find_executable('celery'))
    try:
        clry_main(clry_args)
    except OpenKongqiError as e:
        sys.stderr.write(e.message + "\n")
        sys.exit(1)


if __name__ == '__main__':          # pragma: no cover
    okq_server()



# -*- coding: utf-8 -*-

from __future__ import absolute_import, print_function, unicode_literals

from .conf import settings
from .exceptions import UUIDNotFoundError
from .utils import get_uuid, load_tree, passthrough_loader

_STATIONS_MAP = load_tree(settings['STATIONS_MAP_DIR'],
                          data_loader=passthrough_loader)


def get_station_map(uuid=None):
    """Get the entire station map, given a UUID

    Usage::
        >>> from openkongqi.stations import get_station_map
        >>> get_uuid_map('cn-shanghai')
        ...

    :param uuid: a UUID key
    :type uuid: str
    """
    try:
        id_map = _STATIONS_MAP[uuid]
    except KeyError:
        raise UUIDNotFoundError("Unknown UUID ({})".format(uuid))
    return id_map


def get_all_uuids():
    """Return a sorted list of complete station UUIDs."""
    res = []
    for country, station_map in _STATIONS_MAP.items():
        res.extend([get_uuid(country, s) for s in station_map.values()])
    return sorted(res)
